@@ -73,7 +73,9 @@ typedef enum {
         CHOOSE_LANGUAGE,
         CHOOSE_REGION,
         ADD_INPUT,
-        REMOVE_INPUT
+        REMOVE_INPUT,
+        MOVE_UP_INPUT,
+        MOVE_DOWN_INPUT,
 } SystemOp;
 
 struct _CcRegionPanelPrivate {
@@ -114,6 +116,8 @@ struct _CcRegionPanelPrivate {
         GtkWidget *input_list;
         GtkWidget *add_input;
         GtkWidget *remove_input;
+        GtkWidget *move_up_input;
+        GtkWidget *move_down_input;
         GtkWidget *show_config;
         GtkWidget *show_layout;
 
@@ -228,7 +232,7 @@ show_restart_notification (CcRegionPanel *self,
         GtkWidget *box;
         GtkWidget *label;
         GtkWidget *button;
-        gchar *current_locale;
+        gchar *current_locale = NULL;
 
         if (priv->notification)
                 return;
@@ -482,6 +486,8 @@ show_region_chooser (CcRegionPanel *self,
 
 static void show_input_chooser (CcRegionPanel *self);
 static void remove_selected_input (CcRegionPanel *self);
+static void move_selected_input (CcRegionPanel *self,
+                                 SystemOp       op);
 
 static void
 permission_acquired (GObject      *source,
@@ -515,6 +521,10 @@ permission_acquired (GObject      *source,
                 case REMOVE_INPUT:
                         remove_selected_input (self);
                         break;
+                case MOVE_UP_INPUT:
+                case MOVE_DOWN_INPUT:
+                        move_selected_input (self, priv->op);
+                        break;
                 default:
                         g_warning ("Unknown privileged operation: %d\n", priv->op);
                         break;
@@ -622,13 +632,14 @@ static void
 update_language_from_user (CcRegionPanel *self)
 {
 	CcRegionPanelPrivate *priv = self->priv;
-        const gchar *language;
+        const gchar *language = NULL;
 
 #ifndef __CYGWIN__
         if (act_user_is_loaded (priv->user))
                 language = act_user_get_language (priv->user);
-        else
 #endif
+
+        if (language == NULL || *language == '\0')
                 language = setlocale (LC_MESSAGES, NULL);
 
         g_free (priv->language);
@@ -840,7 +851,7 @@ add_input_row (CcRegionPanel   *self,
         box = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
         gtk_container_add (GTK_CONTAINER (row), box);
         label = gtk_label_new (name);
-        gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
+        gtk_widget_set_halign (label, GTK_ALIGN_START);
         gtk_widget_set_margin_start (label, 20);
         gtk_widget_set_margin_end (label, 20);
         gtk_widget_set_margin_top (label, 6);
@@ -1002,10 +1013,10 @@ update_buttons (CcRegionPanel *self)
 	CcRegionPanelPrivate *priv = self->priv;
         GtkListBoxRow *selected;
         GList *children;
-        gboolean multiple_sources;
+        guint n_rows;
 
         children = gtk_container_get_children (GTK_CONTAINER (priv->input_list));
-        multiple_sources = g_list_next (children) != NULL;
+        n_rows = g_list_length (children);
         g_list_free (children);
 
         selected = gtk_list_box_get_selected_row (GTK_LIST_BOX (priv->input_list));
@@ -1013,6 +1024,8 @@ update_buttons (CcRegionPanel *self)
                 gtk_widget_set_visible (priv->show_config, FALSE);
                 gtk_widget_set_sensitive (priv->remove_input, FALSE);
                 gtk_widget_set_sensitive (priv->show_layout, FALSE);
+                gtk_widget_set_sensitive (priv->move_up_input, FALSE);
+                gtk_widget_set_sensitive (priv->move_down_input, FALSE);
         } else {
                 GDesktopAppInfo *app_info;
 
@@ -1020,11 +1033,13 @@ update_buttons (CcRegionPanel *self)
 
                 gtk_widget_set_visible (priv->show_config, app_info != NULL);
                 gtk_widget_set_sensitive (priv->show_layout, TRUE);
-                gtk_widget_set_sensitive (priv->remove_input, multiple_sources);
+                gtk_widget_set_sensitive (priv->remove_input, n_rows > 1);
+                gtk_widget_set_sensitive (priv->move_up_input, gtk_list_box_row_get_index (selected) > 0);
+                gtk_widget_set_sensitive (priv->move_down_input, gtk_list_box_row_get_index (selected) < n_rows - 1);
         }
 
         gtk_widget_set_visible (priv->options_button,
-                                multiple_sources && !priv->login);
+                                n_rows > 1 && !priv->login);
 }
 
 static void
@@ -1092,31 +1107,6 @@ update_input (CcRegionPanel *self)
         }
 }
 
-static void
-apologize_for_no_ibus_login (CcRegionPanel *self)
-{
-        GtkWidget *dialog;
-        GtkWidget *toplevel;
-        GtkWidget *image;
-
-        toplevel = gtk_widget_get_toplevel (GTK_WIDGET (self));
-
-        dialog = gtk_message_dialog_new (GTK_WINDOW (toplevel),
-                                         GTK_DIALOG_MODAL | GTK_DIALOG_USE_HEADER_BAR,
-                                         GTK_MESSAGE_OTHER,
-                                         GTK_BUTTONS_OK,
-                                         _("Sorry"));
-        gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
-                                                  "%s", _("Input methods can't be used on the login screen"));
-        image = gtk_image_new_from_icon_name ("face-sad-symbolic",
-                                              GTK_ICON_SIZE_DIALOG);
-        gtk_widget_show (image);
-        gtk_message_dialog_set_image (GTK_MESSAGE_DIALOG (dialog), image);
-
-        gtk_dialog_run (GTK_DIALOG (dialog));
-        gtk_widget_destroy (dialog);
-}
-
 static gboolean
 input_source_already_added (CcRegionPanel *self,
                             const gchar   *id)
@@ -1158,13 +1148,10 @@ run_input_chooser (CcRegionPanel *self, GtkWidget *chooser)
                                 type = INPUT_SOURCE_TYPE_XKB;
                         }
 
-                        if (self->priv->login && g_str_equal (type, INPUT_SOURCE_TYPE_IBUS)) {
-                                apologize_for_no_ibus_login (self);
-                        } else {
-                                add_input_row (self, type, id, name, app_info);
-                                update_buttons (self);
-                                update_input (self);
-                        }
+                        add_input_row (self, type, id, name, app_info);
+                        update_buttons (self);
+                        update_input (self);
+
                         g_free (id);
                         g_free (name);
                         g_clear_object (&app_info);
@@ -1185,6 +1172,7 @@ show_input_chooser (CcRegionPanel *self)
         if (!chooser) {
                 toplevel = gtk_widget_get_toplevel (GTK_WIDGET (self));
                 chooser = cc_input_chooser_new (GTK_WINDOW (toplevel),
+                                                priv->login,
                                                 priv->xkb_info,
 #ifdef HAVE_IBUS
                                                 priv->ibus_engines
@@ -1289,6 +1277,71 @@ remove_selected_input (CcRegionPanel *self)
         }
 }
 
+static void
+do_move_selected_input (CcRegionPanel *self,
+                        SystemOp       op)
+{
+	CcRegionPanelPrivate *priv = self->priv;
+        GtkListBoxRow *selected;
+        gint idx;
+
+        g_assert (op == MOVE_UP_INPUT || op == MOVE_DOWN_INPUT);
+
+        selected = gtk_list_box_get_selected_row (GTK_LIST_BOX (priv->input_list));
+        g_assert (selected);
+
+        idx = gtk_list_box_row_get_index (selected);
+        if (op == MOVE_UP_INPUT)
+                idx -= 1;
+        else
+                idx += 1;
+
+        gtk_list_box_unselect_row (GTK_LIST_BOX (priv->input_list), selected);
+
+        g_object_ref (selected);
+        gtk_container_remove (GTK_CONTAINER (priv->input_list), GTK_WIDGET (selected));
+        gtk_list_box_insert (GTK_LIST_BOX (priv->input_list), GTK_WIDGET (selected), idx);
+        g_object_unref (selected);
+
+        gtk_list_box_select_row (GTK_LIST_BOX (priv->input_list), selected);
+
+        cc_list_box_adjust_scrolling (GTK_LIST_BOX (priv->input_list));
+
+        update_buttons (self);
+        update_input (self);
+}
+
+static void
+move_selected_input (CcRegionPanel *self,
+                     SystemOp       op)
+{
+	CcRegionPanelPrivate *priv = self->priv;
+
+        if (!priv->login) {
+                do_move_selected_input (self, op);
+        } else if (g_permission_get_allowed (priv->permission)) {
+                do_move_selected_input (self, op);
+        } else if (g_permission_get_can_acquire (priv->permission)) {
+                priv->op = op;
+                g_permission_acquire_async (priv->permission,
+                                            NULL,
+                                            permission_acquired,
+                                            self);
+        }
+}
+
+static void
+move_selected_input_up (CcRegionPanel *self)
+{
+        move_selected_input (self, MOVE_UP_INPUT);
+}
+
+static void
+move_selected_input_down (CcRegionPanel *self)
+{
+        move_selected_input (self, MOVE_DOWN_INPUT);
+}
+
 static void
 show_selected_settings (CcRegionPanel *self)
 {
@@ -1430,6 +1483,8 @@ setup_input_section (CcRegionPanel *self)
         priv->input_list = WID ("input_list");
         priv->add_input = WID ("input_source_add");
         priv->remove_input = WID ("input_source_remove");
+        priv->move_up_input = WID ("input_source_up");
+        priv->move_down_input = WID ("input_source_down");
         priv->show_config = WID ("input_source_config");
         priv->show_layout = WID ("input_source_layout");
 
@@ -1439,6 +1494,10 @@ setup_input_section (CcRegionPanel *self)
                                   G_CALLBACK (add_input), self);
         g_signal_connect_swapped (priv->remove_input, "clicked",
                                   G_CALLBACK (remove_selected_input), self);
+        g_signal_connect_swapped (priv->move_up_input, "clicked",
+                                  G_CALLBACK (move_selected_input_up), self);
+        g_signal_connect_swapped (priv->move_down_input, "clicked",
+                                  G_CALLBACK (move_selected_input_down), self);
         g_signal_connect_swapped (priv->show_config, "clicked",
                                   G_CALLBACK (show_selected_settings), self);
         g_signal_connect_swapped (priv->show_layout, "clicked",
